package lxc

import (
	"fmt"
	"io"
	"io/ioutil"
	"os"
	"path/filepath"
	"regexp"
	"strings"

	"launchpad.net/juju-core/environs/agent"
	"launchpad.net/juju-core/log"
	"launchpad.net/juju-core/state"
	"launchpad.net/juju-core/upstart"
	"launchpad.net/juju-core/version"
	"launchpad.net/juju-core/worker/deployer"
)

// LxcContext stores details of the LXC environment on the local system.
type LxcContext struct {
	// environName identifies the environment the context is running in.
	environName string

	// environSeries identifies the series of the environment.
	environSeries string

	// addresser is used to get the current state server addresses at the time
	// the given unit is deployed.
	addresser deployer.Addresser

	// caCert holds the CA certificate that will be used
	// to validate the state server's certificate, in PEM format.
	caCert []byte

	// deployerTag identifies the agent on whose behalf this context is running.
	deployerTag string

	// initDir specifies the directory used by upstart on the local system.
	// It is typically set to "/etc/init".
	initDir string

	// dataDir specifies the directory used by juju to store its state. It
	// is typically set to "/var/lib/juju".
	dataDir string

	// hostDataDir specifies the directory used by juju on the host system to 
	// store its tools. It is typically set to "/var/lib/juju".
	hostDataDir string

	// logDir specifies the directory to which installed units will write
	// their log files. It is typically set to "/var/log/juju".
	logDir string

	// lxcContainerDir specifies the base directory for all containers. It
	// is typically set to "/var/lib/lxc".
	lxcContainerDir string

	// lxcAutoDir specifies the directory for the links to confguration files
	// needed for autostarting containers. It is typically set to "/etc/lxc/auto".
	lxcAutoDir string
}

var _ deployer.Context = (*LxcContext)(nil)

// NewLxcContext returns a new LXC context, acting on behalf of the
// entity specified in info, that deploys unit agents as upstart jobs in
// "/etc/init" logging to "/var/log/juju". Paths to which agents and tools
// are installed are relative to dataDir; if dataDir or hostDataDir are empty, 
// they will be set to "/var/lib/juju".
func NewLxcContext(environName, environSeries, dataDir, hostDataDir string, caCert []byte,
	deployerTag string, addresser deployer.Addresser) *LxcContext {
	if dataDir == "" {
		dataDir = "/var/lib/juju"
	}
	if hostDataDir == "" {
		hostDataDir = "/var/lib/juju"
	}
	return &LxcContext{
		environName:     environName,
		environSeries:   environSeries,
		dataDir:         dataDir,
		hostDataDir:     hostDataDir,
		addresser:       addresser,
		caCert:          caCert,
		deployerTag:     deployerTag,
		initDir:         "/etc/init",
		logDir:          "/var/log/juju",
		lxcContainerDir: "/var/lib/lxc",
		lxcAutoDir:      "/etc/lxc/auto",
	}
}

// DeployUnit creates a container for a unit and installs the required 
// upstart job in it.
func (ctx *LxcContext) DeployUnit(unitName, initialPassword string) error {
	if ctx.configLinkExists(unitName) {
		return fmt.Errorf("unit %q is already deployed", unitName)
	}

	// Create container and initialize it.
	containerName := ctx.containerName(unitName)
	unitTag := state.UnitTag(unitName)
	if err := ctx.createContainer(containerName); err != nil {
		return err
	}
	if err := ctx.initContainer(containerName, unitName, unitTag, initialPassword); err != nil {
		if e := ctx.destroyContainer(containerName); e != nil {
			return fmt.Errorf("cannot destroy container after init error '%v': %v", err, e)
		}
		return err
	}
	if err := ctx.createConfigLink(unitName); err != nil {
		return fmt.Errorf("cannot create auto start link: %v", err)
	}
	return startContainer(containerName)
}

// RecallUnit destroys the container and removes the link in the
// /etc/lxc/auto folder.
func (ctx *LxcContext) RecallUnit(unitName string) error {
	if !ctx.configLinkExists(unitName) {
		return fmt.Errorf("unit %q is not deployed", unitName)
	}
	if err := ctx.destroyConfigLink(unitName); err != nil {
		return err
	}

	// Destroy container.
	containerName := ctx.containerName(unitName)
	return ctx.destroyContainer(containerName)
}

var autoRe = regexp.MustCompile("^juju-([a-z0-9-]+):([a-z0-9-]+):unit-([a-z0-9-]+)-([0-9]+)\\.conf$")

// DeployedUnits returns which units are deployed right now.
func (ctx *LxcContext) DeployedUnits() ([]string, error) {
	fis, err := ioutil.ReadDir(ctx.lxcAutoDir)
	if err != nil {
		return nil, err
	}
	var installed []string
	for _, fi := range fis {
		groups := autoRe.FindStringSubmatch(fi.Name())
		if len(groups) == 5 {
			if groups[2] != ctx.deployerTag {
				continue
			}
			unitName := groups[3] + "/" + groups[4]
			if !state.IsUnitName(unitName) {
				continue
			}
			installed = append(installed, unitName)
		}
	}
	return installed, nil
}

// upstartService returns an upstart.Service corresponding to the specified
// unit. Its name is badged according to the entity responsible for the
// context, so as to distinguish its own jobs from those installed by other
// means.
func (ctx *LxcContext) upstartService(unitName string) *upstart.Service {
	unitTag := state.UnitTag(unitName)
	svcName := "jujud-" + ctx.deployerTag + ":" + unitTag
	svc := upstart.NewService(svcName)
	svc.InitDir = ctx.initDir
	return svc
}

// createContainer creates a new LXC container with the given name.
func (ctx *LxcContext) createContainer(containerName string) error {
	if ctx.containerExists(containerName) {
		return fmt.Errorf("container %q is already created", containerName)
	}
	return createContainer(containerName, ctx.environSeries)
}

// initContainer creates the needed directories inside the container
// and copies the tools.
func (ctx *LxcContext) initContainer(containerName, unitName, unitTag, initialPassword string) error {
	hostToolsDir := agent.SharedToolsDir(ctx.hostDataDir, version.Current)
	toolsDir := ctx.containerDir(containerName, agent.ToolsDir(ctx.dataDir, unitTag))
	err := os.MkdirAll(toolsDir, 0777)
	if err != nil {
		return err
	}
	defer removeOnErr(&err, toolsDir)

	agentsDir := ctx.containerDir(containerName, ctx.dataDir, "agents", unitTag)
	err = os.MkdirAll(agentsDir, 0777)
	if err != nil {
		return err
	}
	defer removeOnErr(&err, agentsDir)

	logDir := ctx.containerDir(containerName, ctx.logDir)
	err = os.MkdirAll(ctx.containerDir(containerName, ctx.logDir), 0777)
	if err != nil {
		return err
	}
	defer removeOnErr(&err, logDir)

	if err = copyTools(hostToolsDir, toolsDir); err != nil {
		return err
	}
	if _, err = agent.ChangeAgentTools(ctx.hostDataDir, unitTag, version.Current); err != nil {
		return err
	}

	// Prepare the agent's configuration data.
	info := state.Info{
		Addrs:  ctx.addresser.Addresses(),
		Tag:    unitTag,
		CACert: ctx.caCert,
	}
	confDataDir := ctx.containerDir(containerName, ctx.dataDir)
	conf := &agent.Conf{
		DataDir:     confDataDir,
		OldPassword: initialPassword,
		StateInfo:   &info,
	}
	conf.StateInfo.Tag = unitTag
	conf.StateInfo.Password = ""
	if err = conf.Write(); err != nil {
		return err
	}
	defer removeOnErr(&err, conf.Dir())

	// Install an upstart job that runs the unit agent.
	jujudPath := filepath.Join(toolsDir, "jujud")
	logPath := filepath.Join(ctx.logDir, unitTag+".log")
	cmd := strings.Join([]string{
		jujudPath, "unit",
		"--data-dir", ctx.dataDir,
		"--unit-name", unitName,
		"--debug", // TODO: propagate debug state sensibly
	}, " ")
	svc := ctx.upstartService(unitName)
	uconf := &upstart.Conf{
		Service: *svc,
		Desc:    "juju unit agent for " + unitName,
		Cmd:     cmd,
		Out:     logPath,
	}
	return uconf.Install()
}

// destroyContainer stops and removes the container.
func (ctx *LxcContext) destroyContainer(containerName string) error {
	if !ctx.containerExists(containerName) {
		return fmt.Errorf("container %q does not exist", containerName)
	}
	if err := stopContainer(containerName); err != nil {
		return err
	}
	if err := waitContainerStopped(containerName); err != nil {
		return err
	}
	return destroyContainer(containerName)
}

// containerName returns a unique name for the container and its upstart 
// script based on the unit name.
func (ctx *LxcContext) containerName(unitName string) string {
	unitTag := state.UnitTag(unitName)
	return "juju-" + ctx.environName + ":" + ctx.deployerTag + ":" + unitTag
}

// configLinkExists checks if the configuration link for an automatic start exists.
func (ctx *LxcContext) configLinkExists(unitName string) bool {
	containerName := ctx.containerName(unitName)
	autoConfigPath := filepath.Join(ctx.lxcAutoDir, containerName+".conf")
	containerConfigPath := filepath.Join(ctx.lxcContainerDir, containerName, "config")
	targetPath, err := os.Readlink(autoConfigPath)
	return err == nil && targetPath == containerConfigPath
}

// createConfigLink links from a container config to the LXC auto-start folder.
func (ctx *LxcContext) createConfigLink(unitName string) error {
	containerName := ctx.containerName(unitName)
	autoConfigPath := filepath.Join(ctx.lxcAutoDir, containerName+".conf")
	containerConfigPath := filepath.Join(ctx.lxcContainerDir, containerName, "config")
	return os.Symlink(containerConfigPath, autoConfigPath)
}

// destroyConfigLink removes the link from a container config to the LXC auto-start folder.
func (ctx *LxcContext) destroyConfigLink(unitName string) error {
	containerName := ctx.containerName(unitName)
	autoConfigPath := filepath.Join(ctx.lxcAutoDir, containerName+".conf")
	return os.Remove(autoConfigPath)
}

// containerDir returns the full qualified name of a directory
// inside the container.
func (ctx *LxcContext) containerDir(containerName string, dirs ...string) string {
	rootfs := []string{ctx.lxcContainerDir, containerName, "rootfs"}
	return filepath.Join(append(rootfs, dirs...)...)
}

// containerExists checks if the container is created.
func (ctx *LxcContext) containerExists(containerName string) bool {
	fi, err := os.Stat(ctx.containerDir(containerName))
	if err != nil {
		if os.IsNotExist(err) {
			return false
		}
		panic(err)
	}
	return fi.IsDir()
}

// copyTools copies the hosts juju tools into the container.
func copyTools(hostToolsDir, containerToolsDir string) error {
	hostLen := len(hostToolsDir)
	copyTool := func(hostPath string, info os.FileInfo, err error) error {
		if err != nil {
			return err
		}
		containerPath := filepath.Join(containerToolsDir, hostPath[hostLen:])
		if info.IsDir() {
			// Create directory.
			if err := os.MkdirAll(containerPath, info.Mode().Perm()); err != nil {
				return err
			}
		} else {
			// Copy file.
			sourceFile, err := os.Open(hostPath)
			if err != nil {
				return err
			}
			defer sourceFile.Close()
			destFile, err := os.Create(containerPath)
			if err != nil {
				return err
			}
			defer destFile.Close()
			if _, err := io.Copy(destFile, sourceFile); err != nil {
				return err
			}
			return destFile.Chmod(info.Mode())
		}
		return nil
	}
	return filepath.Walk(hostToolsDir, copyTool)
}

// removeOnErr removes a path in case of an error.
func removeOnErr(err *error, path string) {
	if *err != nil {
		if e := os.Remove(path); e != nil {
			log.Warningf("installer: cannot remove %q: %v", path, e)
		}
	}
}
