package lxc

import (
	"fmt"
	"io/ioutil"
	"os/exec"
	"regexp"
	"strings"
)

const (
	defaultAddr   = "10.0.3.1"
	defaultBridge = "lxcbr0"
)

var (
	confPath string         = "/etc/default/lxc"
	addrRE   *regexp.Regexp = regexp.MustCompile(`\n\s*LXC_ADDR="(\d+.\d+.\d+.\d+)"`)
	bridgeRE *regexp.Regexp = regexp.MustCompile(`\n\s*LXC_BRIDGE="(\w+)"`)
)

// Error reports the failure of a LXC command.
type Error struct {
	Name   string
	Err    error
	Output []string
}

func (e Error) Error() string {
	if e.Output == nil {
		return fmt.Sprintf("error executing %q: %v", e.Name, e.Err)
	}
	if len(e.Output) == 1 {
		return fmt.Sprintf("error executing %q: %v", e.Name, e.Output[0])
	}
	return fmt.Sprintf("error executing %q: %s", e.Name, strings.Join(e.Output, "; "))
}

// createContainer creates a new ubuntu cloud container.
func createContainer(containerName, series string) error {
	hostId := strings.Replace(containerName, ":", "-", -1)
	_, err := run("lxc-create",
		"-n", containerName,
		"-t", "ubuntu-cloud",
		"--",
		"-d",
		"-i", hostId,
		"-r", series)
	return err
}

// startContainer starts a container.
func startContainer(containerName string) error {
	_, err := run("lxc-start", "-n", containerName)
	return err
}

// stopContainer stops a container.
func stopContainer(containerName string) error {
	_, err := run("lxc-stop", "-n", containerName)
	return err
}

// waitContainerStopped waits until a container stopped.
func waitContainerStopped(containerName string) error {
	_, err := run("lxc-wait", "-n", containerName, "-s", "STOPPED")
	return err
}

// destroyContainer destroys a container.
func destroyContainer(containerName string) error {
	_, err := run("lxc-destroy", "-n", containerName)
	return err
}

// startNetwork starts the lxc network subsystem.
func startNetwork() error {
	goal, _, err := networkStatus()
	if err != nil {
		return err
	}
	if goal != "start" {
		_, err := run("start", "lxc-net")
		if err != nil {
			return err
		}
	}
	return nil
}

// stopNetwork stops the lxc network subsystem.
func stopNetwork() error {
	goal, _, err := networkStatus()
	if err != nil {
		return err
	}
	if goal != "stop" {
		_, err := run("stop", "lxc-net")
		if err != nil {
			return err
		}
	}
	return nil
}

// isNotworkRunning checks if the lxc network subsystem
// is running.
func isNetworkRunning() (bool, error) {
	_, status, err := networkStatus()
	if err != nil {
		return false, err
	}
	return status == "running", nil
}

// networkAttributes returns the lxc network attributes: 
// starting IP address and bridge name.
func networkAttributes() (addr, bridge string, err error) {
	config, err := readConf()
	if err != nil {
		return "", "", err
	}
	addr = config["address"]
	if addr == "" {
		addr = defaultAddr
	}
	bridge = config["bridge"]
	if bridge == "" {
		bridge = defaultBridge
	}
	return addr, bridge, nil
}

// networkStatus returns the status of the lxc network subsystem.
func networkStatus() (goal, status string, err error) {
	output, err := run("status", "lxc-net")
	if err != nil {
		return "", "", err
	}
	fields := strings.Fields(output)
	if len(fields) != 2 {
		return "", "", fmt.Errorf("unexpected status output: %q", output)
	}
	fields = strings.Split(fields[1], "/")
	if len(fields) != 2 {
		return "", "", fmt.Errorf("unexpected status output: %q", output)
	}
	return fields[0], fields[1], nil
}

// run executes the passed command and returns the out.
func run(name string, args ...string) (string, error) {
	cmd := exec.Command(name, args...)
	// LXC tools do not use stdout and stderr in a predictable
	// way; based on experimentation, the most convenient
	// solution is to combine them and leave the client to
	// determine sanity as best it can.
	out, err := cmd.CombinedOutput()
	if err != nil {
		return "", runError(name, err, out)
	}
	return string(out), nil
}

// runError creates an error if run fails.
func runError(name string, err error, out []byte) error {
	e := &Error{name, err, nil}
	for _, l := range strings.Split(string(out), "\n") {
		if strings.HasPrefix(l, name+": ") {
			// LXC tools do not always print their output with
			// the command name as prefix. The name is part of
			// the error struct, so stip it from the output if
			// printed.
			l = l[len(name)+2:]
		}
		if l != "" {
			e.Output = append(e.Output, l)
		}
	}
	return e
}

// keyValues retrieves key/value pairs out of a command out.
func keyValues(raw string, sep string) map[string]string {
	kv := map[string]string{}
	lines := strings.Split(raw, "\n")
	for _, line := range lines {
		parts := strings.SplitN(line, sep, 2)
		if len(parts) == 2 {
			kv[parts[0]] = strings.TrimSpace(parts[1])
		}
	}
	return kv
}

// readConf reads the LXC network address and bridge interface
// out of the configuration file /etc/default/lxc.
func readConf() (map[string]string, error) {
	conf := make(map[string]string)
	confData, err := ioutil.ReadFile(confPath)
	if err != nil {
		return nil, err
	}
	fetchValue := func(field string, re *regexp.Regexp) {
		groups := re.FindStringSubmatch(string(confData))
		if len(groups) == 2 {
			conf[field] = groups[1]
		}
	}
	fetchValue("address", addrRE)
	fetchValue("bridge", bridgeRE)
	return conf, nil
}
