package lxc_test

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"os"
	"path/filepath"
	stdtesting "testing"
	"text/template"

	. "launchpad.net/gocheck"
	"launchpad.net/juju-core/environs/agent"
	"launchpad.net/juju-core/state"
	coretesting "launchpad.net/juju-core/testing"
	"launchpad.net/juju-core/version"
	"launchpad.net/juju-core/worker/deployer/lxc"
)

func TestPackage(t *stdtesting.T) {
	coretesting.MgoTestPackage(t)
}

type LxcContextSuite struct {
	LxcToolsFixture
}

var _ = Suite(&LxcContextSuite{})

func (s *LxcContextSuite) SetUpTest(c *C) {
	s.LxcToolsFixture.SetUp(c)
}

func (s *LxcContextSuite) TearDownTest(c *C) {
	s.LxcToolsFixture.TearDown(c)
}

func (s *LxcContextSuite) TestDeployRecall(c *C) {
	ctx0 := s.getContext(c, "test-entity-0")
	units, err := ctx0.DeployedUnits()
	c.Assert(err, IsNil)
	c.Assert(units, HasLen, 0)
	s.assertUpstartCount(c, 0)

	err = ctx0.DeployUnit("foo/123", "some-password")
	c.Assert(err, IsNil)
	units, err = ctx0.DeployedUnits()
	c.Assert(err, IsNil)
	c.Assert(units, DeepEquals, []string{"foo/123"})
	s.assertUpstartCount(c, 1)
	s.checkUnitInstalled(c, "test-entity-0", "foo/123", "some-password")

	err = ctx0.RecallUnit("foo/123")
	c.Assert(err, IsNil)
	units, err = ctx0.DeployedUnits()
	c.Assert(err, IsNil)
	c.Assert(units, HasLen, 0)
	s.assertUpstartCount(c, 0)
	s.checkUnitRemoved(c, "test-entity-0", "foo/123")
}

func (s *LxcContextSuite) TestIndependentManagers(c *C) {
	ctx0 := s.getContext(c, "test-entity-0")
	err := ctx0.DeployUnit("foo/123", "some-password")
	c.Assert(err, IsNil)

	ctx1 := s.getContext(c, "test-entity-1")
	units, err := ctx1.DeployedUnits()
	c.Assert(err, IsNil)
	c.Assert(units, HasLen, 0)

	err = ctx1.RecallUnit("foo/123")
	c.Assert(err, ErrorMatches, `unit "foo/123" is not deployed`)
	s.checkUnitInstalled(c, "test-entity-0", "foo/123", "some-password")

	units, err = ctx0.DeployedUnits()
	c.Assert(err, IsNil)
	c.Assert(units, DeepEquals, []string{"foo/123"})
	s.assertUpstartCount(c, 1)
	s.checkUnitInstalled(c, "test-entity-0", "foo/123", "some-password")
}

type LxcToolsFixture struct {
	origPath        string
	hostDataDir     string
	initDir         string
	logDir          string
	lxcContainerDir string
	lxcAutoDir      string
	binDir          string
}

var fakeJujud = "#!/bin/bash\n# fake-jujud\nexit 0\n"

func (fix *LxcToolsFixture) SetUp(c *C) {
	// Create some needed temporary directory and manipulate the path
	// for the fake binaries.
	fix.initDir = c.MkDir()
	fix.logDir = c.MkDir()
	fix.hostDataDir = c.MkDir()
	fix.lxcContainerDir = c.MkDir()
	fix.lxcAutoDir = c.MkDir()
	fix.binDir = c.MkDir()
	fix.origPath = os.Getenv("PATH")
	os.Setenv("PATH", fix.binDir+":"+fix.origPath)
	// Create the faked tools and binaries.
	toolsDir := agent.SharedToolsDir(fix.hostDataDir, version.Current)
	err := os.MkdirAll(toolsDir, 0755)
	c.Assert(err, IsNil)
	jujudPath := filepath.Join(toolsDir, "jujud")
	err = ioutil.WriteFile(jujudPath, []byte(fakeJujud), 0755)
	c.Assert(err, IsNil)
	urlPath := filepath.Join(toolsDir, "downloaded-url.txt")
	err = ioutil.WriteFile(urlPath, []byte("http://example.com/tools"), 0644)
	c.Assert(err, IsNil)
	fix.makeLxcBins(c)
	fix.makeBin(c, "status", `echo "blah stop/waiting"`)
	fix.makeBin(c, "stopped-status", `echo "blah stop/waiting"`)
	fix.makeBin(c, "started-status", `echo "blah start/running, process 666"`)
	fix.makeBin(c, "start", "cp $(which started-status) $(which status)")
	fix.makeBin(c, "stop", "cp $(which stopped-status) $(which status)")
}

func (fix *LxcToolsFixture) TearDown(c *C) {
	os.Setenv("PATH", fix.origPath)
}

var LxcBins = []struct {
	Name string
	Code string
}{
	{
		Name: "lxc-create",
		Code: `#!/bin/bash
# lxc-create
mkdir -p {{.LxcContainerDir}}/$2/rootfs
touch {{.LxcContainerDir}}/$2/config
exit 0`,
	}, {
		Name: "lxc-start",
		Code: `#!/bin/bash
# lxc-start
exit 0`,
	}, {
		Name: "lxc-stop",
		Code: `#!/bin/bash
# lxc-stop
exit 0`,
	}, {
		Name: "lxc-wait",
		Code: `#!/bin/bash
# lxc-wait
exit 0`,
	}, {
		Name: "lxc-destroy",
		Code: `#!/bin/bash
# lxc-destroy
rm -R {{.LxcContainerDir}}/$2
exit 0`,
	},
}

func (fix *LxcToolsFixture) makeLxcBins(c *C) {
	conf := struct {
		LxcContainerDir string
		LxcAutoDir      string
	}{
		LxcContainerDir: fix.lxcContainerDir,
		LxcAutoDir:      fix.lxcAutoDir,
	}
	for _, lxcBin := range LxcBins {
		var buf bytes.Buffer
		binT := template.Must(template.New("").Parse(lxcBin.Code))
		err := binT.Execute(&buf, conf)
		c.Assert(err, IsNil)
		binPath := filepath.Join(fix.binDir, lxcBin.Name)
		err = ioutil.WriteFile(binPath, buf.Bytes(), 0755)
		c.Assert(err, IsNil)
	}
}

func (fix *LxcToolsFixture) makeBin(c *C, name, script string) {
	path := filepath.Join(fix.binDir, name)
	err := ioutil.WriteFile(path, []byte("#!/bin/bash\n"+script), 0755)
	c.Assert(err, IsNil)
}

func (fix *LxcToolsFixture) assertUpstartCount(c *C, count int) {
	fis, err := ioutil.ReadDir(fix.lxcAutoDir)
	c.Assert(err, IsNil)
	c.Assert(fis, HasLen, count)
}

func (fix *LxcToolsFixture) getContext(c *C, deployerTag string) *lxc.LxcContext {
	return lxc.NewTestLxcContext(deployerTag, fix.initDir, fix.hostDataDir, fix.lxcContainerDir, fix.lxcAutoDir)
}

func (fix *LxcToolsFixture) checkUnitInstalled(c *C, dname, uname, password string) {
	unitTag := state.UnitTag(uname)
	lxcContainerName := "juju-lxc-test:" + dname + ":" + unitTag
	lxcContainerPath := filepath.Join(fix.lxcContainerDir, lxcContainerName)
	lxcConfName := fmt.Sprintf("%s.conf", lxcContainerName)
	lxcConfPath := filepath.Join(fix.lxcAutoDir, lxcConfName)

	_, err := os.Stat(lxcContainerPath)
	c.Assert(err, IsNil)
	_, err = os.Stat(lxcConfPath)
	c.Assert(err, IsNil)

	dataDir := filepath.Join(fix.lxcContainerDir, lxcContainerName, "rootfs/var/lib/juju")
	conf, err := agent.ReadConf(dataDir, unitTag)
	c.Assert(err, IsNil)
	c.Assert(conf, DeepEquals, &agent.Conf{
		DataDir:     dataDir,
		OldPassword: password,
		StateInfo: &state.Info{
			Addrs:  []string{"s1:123", "s2:123"},
			CACert: []byte("test-cert"),
			Tag:    unitTag,
		},
	})

	toolsDir := agent.ToolsDir(dataDir, unitTag)
	jujudPath := filepath.Join(toolsDir, "jujud")
	jujudData, err := ioutil.ReadFile(jujudPath)
	c.Assert(err, IsNil)
	c.Assert(string(jujudData), Equals, fakeJujud)
}

func (fix *LxcToolsFixture) checkUnitRemoved(c *C, ename, uname string) {
	unitTag := state.UnitTag(uname)
	lxcContainerName := "juju-lxc-test:" + ename + ":" + unitTag
	lxcContainerPath := filepath.Join(fix.lxcContainerDir, lxcContainerName)
	lxcConfName := fmt.Sprintf("%s.conf", lxcContainerName)
	lxcConfPath := filepath.Join(fix.lxcAutoDir, lxcConfName)

	_, err := os.Stat(lxcContainerPath)
	c.Assert(os.IsNotExist(err), Equals, true)
	_, err = os.Stat(lxcConfPath)
	c.Assert(os.IsNotExist(err), Equals, true)
}
